from rest_framework import filters, pagination, serializers

try:
    from django_filters.rest_framework import DjangoFilterBackend
except ImportError:
    # Older versions of DRF and django_filters
    from rest_framework.filters import DjangoFilterBackend
from django.core.exceptions import FieldDoesNotExist

try:
    from django.db.models.fields.reverse_related import ManyToOneRel, OneToOneRel
except ImportError:
    # Django 1.8
    from django.db.models.fields.related import ManyToOneRel, OneToOneRel

from django.conf import settings
from django.db.models.fields import NOT_PROVIDED


class NullToDefaultMixin(object):

    def __init__(self, *args, **kwargs):
        super(NullToDefaultMixin, self).__init__(*args, **kwargs)
        for field in self.Meta.fields:
            try:
                model_field = self.Meta.model._meta.get_field(field)
                if hasattr(model_field, 'default') and model_field.default != NOT_PROVIDED:
                    self.fields[field].allow_null = True
            except FieldDoesNotExist:
                pass

    def validate(self, data):
        for field in self.Meta.fields:
            try:
                model_field = self.Meta.model._meta.get_field(field)
                if hasattr(model_field, 'default') and model_field.default != NOT_PROVIDED and \
                        data.get(field, NOT_PROVIDED) is None:
                    data.pop(field)
            except FieldDoesNotExist:
                pass

        return data



def serializer_factory(endpoint):

    from .app_settings import settings

    meta_attrs = {
        'model': endpoint.model,
        'fields': endpoint.get_fields_for_serializer()
    }
    meta_parents = (object, )
    if hasattr(endpoint.base_serializer, 'Meta'):
        meta_parents = (endpoint.base_serializer.Meta, ) + meta_parents

    Meta = type('Meta', meta_parents, meta_attrs)

    cls_name = '{}Serializer'.format(endpoint.model.__name__)
    cls_attrs = {
        'Meta': Meta,
    }

    for meta_field in meta_attrs['fields']:
        try:
            model_field = endpoint.model._meta.get_field(meta_field)
            if isinstance(model_field, OneToOneRel):
                cls_attrs[meta_field] = serializers.PrimaryKeyRelatedField(read_only=True)
            elif isinstance(model_field, ManyToOneRel):
                cls_attrs[meta_field] = serializers.PrimaryKeyRelatedField(many=True, read_only=True)
        except FieldDoesNotExist:
            cls_attrs[meta_field] = serializers.ReadOnlyField()

    return type(cls_name, (NullToDefaultMixin, endpoint.base_serializer, ), cls_attrs)


def viewset_factory(endpoint):

    base_viewset = endpoint.get_base_viewset()

    cls_name = '{}ViewSet'.format(endpoint.model.__name__)
    tmp_cls_attrs = {
        'serializer_class': endpoint.get_serializer(),
        'queryset': endpoint.model.objects.all(),
        'endpoint': endpoint,
        '__doc__': base_viewset.__doc__
    }

    cls_attrs = {
        key: value
        for key, value in tmp_cls_attrs.items() if key == '__doc__' or
        getattr(base_viewset, key, None) is None
    }

    if endpoint.permission_classes is not None:
        cls_attrs['permission_classes'] = endpoint.permission_classes

    filter_backends = getattr(endpoint.get_base_viewset(), 'filter_backends', ())
    if filter_backends is None:
        filter_backends = []
    else:
        filter_backends = list(filter_backends)

    for filter_type, backend in (
        ('filter_fields', DjangoFilterBackend),
        ('search_fields', filters.SearchFilter),
        ('ordering_fields', filters.OrderingFilter),
    ):

        if getattr(endpoint, filter_type, None) is not None:
            filter_backends.append(backend)
            cls_attrs[filter_type] = getattr(endpoint, filter_type)

    if hasattr(endpoint, 'filter_class'):
        if DjangoFilterBackend not in filter_backends:
            filter_backends.append(DjangoFilterBackend)
        cls_attrs['filter_class'] = endpoint.filter_class

    if len(filter_backends) > 0:
        cls_attrs['filter_backends'] = filter_backends

    if endpoint.page_size is not None:

        pg_cls_name = '{}Pagination'.format(endpoint.model.__name__)
        pg_cls_attrs = {
            'page_size': endpoint.page_size,
            'page_size_query_param': 'page_size',
            'max_page_size': settings.REST_FRAMEWORK.get('PAGE_SIZE', 250)
        }
        pg_cls = type(pg_cls_name, (pagination.PageNumberPagination, ), pg_cls_attrs)

        cls_attrs['pagination_class'] = pg_cls

    rv = type(cls_name, (endpoint.get_base_viewset(),), cls_attrs)

    for method_name in dir(endpoint):
        method = getattr(endpoint, method_name)
        if getattr(method, 'action_type', None) in ['custom', 'bulk']:
            setattr(rv, method_name, method)

    return rv



from django.core.management.base import BaseCommand, CommandError
from django.utils.module_loading import import_string
from django.conf import settings as django_settings

from export_app import settings
from export_app.base import SerializerExporterWithFields, ModelNotFoundException


class Command(SerializerExporterWithFields, BaseCommand):
    help = 'Export DRF serializer definition to a frontend model definition'

    def add_arguments(self, parser):
        parser.add_argument('model_endpoint', nargs='*',
                            help="The shorthand url(s) of the endpoint(s) for which you'd like to export models. Eg: 'sample/products'")
        parser.add_argument('--all', default=False, action='store_true',
                            help="Export all models corresponding to endpoints registered with your router")
        parser.add_argument('--adapter_name', default=settings.ADAPTER,
                            help='Defaults to {}'.format(settings.ADAPTER))
        parser.add_argument('--target_app', default=None,
                            help='Force all relationships to the target_app instead of the model\'s app')
        parser.add_argument('--router', default=None,
                            help='Defaults to {}'.format(settings.ROUTER_PATH))
        parser.add_argument('--noinput', default=False, action='store_true')

    def handle(self, *args, **options):
        target_app = options['target_app']
        adapter_name = options['adapter_name']
        if '.' not in adapter_name:
            adapter_name = 'export_app.adapters.{}'.format(adapter_name)
        Adapter = import_string(adapter_name)
        adapter = Adapter()

        if options['router'] is not None:
            self.router = import_string(options['router'])
        endpoints = options['model_endpoint']

        if len(endpoints) == 0 and not options['all']:
            self.print_help('drf_auto_endpoint', 'export')
            return
        elif options['all'] and len(endpoints) > 0:
            raise CommandError('You need to specify either model_endpoint(s) or use --all but you cannot use both at the same time')
        elif options['all']:
            endpoints = getattr(self.router, '_endpoints', {}).keys()

        for endpoint in endpoints:
            print('Exporting {} using {}'.format(endpoint, adapter_name))
            try:
                if adapter.works_with in ['serializer', 'both']:
                    model, serializer_instance, model_name, application_name = \
                        self.get_serializer_for_basename(endpoint)

                    class BogusViewSet(object):
                        pagination_class = None

                    viewset = BogusViewSet()
                if adapter.works_with in ['viewset', 'both']:
                    viewset, model_name, application_name = self.get_viewset_for_basename(endpoint)
            except ModelNotFoundException as e:
                raise CommandError('No viewset found for {}'.format(e.model))

            fields, rels = [], []

            if adapter.requires_fields:
                fields, rels = self.get_fields_for_model(model, serializer_instance, adapter,
                                                         target_app)

            belongsTo = False
            hasMany = False

            for rel in rels:
                if rel['type'] == 'belongsTo':
                    belongsTo = True
                    if hasMany:
                        break
                else:
                    hasMany = True
                    if belongsTo:
                        break

            if adapter.works_with in ['serializer', 'both']:
                context = {
                    'endpoint': endpoint,
                    'model_name': model_name,
                    'application_name': application_name,
                    'fields': fields,
                    'rels': rels,
                    'belongsTo': belongsTo,
                    'hasMany': hasMany,
                    'target_app': target_app,
                    'api_base': settings.BACK_API_BASE,
                }

                adapter.write_to_file(application_name, model_name, context, options['noinput'])
            else:
                adapter.write_to_file(application_name, model_name, viewset, options['noinput'])

        adapter.rebuild_index()



from django.http import Http404
from django.views.generic import TemplateView
from django.utils.module_loading import import_string

from export_app import settings
from export_app.base import SerializerExporterWithFields, ModelNotFoundException


class BaseModelView(SerializerExporterWithFields, TemplateView):
    adapter_class = import_string(settings.ADAPTER)
    content_type = 'text/javascript'

    def get_template_names(self):
        return [self.adapter_class.dynamic_template_name]


class EmberModelView(BaseModelView):
    """
    Generate resource on the fly for Ember
    """

    def get_context_data(self, **kwargs):
        context = super(EmberModelView, self).get_context_data(**kwargs)
        try:
            model, serializer_instance, context['model_name'], context['application_name'] = \
                self.get_serializer_for_basename(self.kwargs['model'])
        except ModelNotFoundException as e:
            raise Http404('No viewset found for {}'.format(e.model))

        for item in ['model_name', 'application_name']:
            context[item] = context[item].replace('_', '-')

        fields, rels = self.get_fields_for_model(model, serializer_instance, self.adapter_class)

        # WIP: for models on the fly, we may need to add url name
        # url_name = settings.URL_NAME
        # if settings.URL_NAMESPACE is not None:
        #     url_name = '{}:{}'.format(settings.URL_NAMESPACE, settings.URL_NAME)

        context['fields'] = fields
        context['rels'] = []
        for rel in rels:
            for item in ['app', 'related_model']:
                rel[item] = rel[item].replace('_', '-')
            context['rels'].append(rel)

        context['ember_app'] = settings.FRONT_APPLICATION_NAME
        return context


class WizardModelView(BaseModelView):

    def get_context_data(self, **kwargs):
        context = super(WizardModelView, self).get_context_data(**kwargs)
        base_name, method_name = self.kwargs['model'].rsplit('/', 1)
        context['application_name'] = 'wizard/{}'.format(base_name.replace('_', '-'))
        context['model_name'] = method_name.replace('_', '-')

        viewset, model_name, application_name = self.get_viewset_for_basename(base_name)
        serializer = getattr(viewset, method_name).serializer
        serializer_instance = serializer()

        fields, rels = self.get_fields_for_model(None, serializer_instance, self.adapter_class)

        context['fields'] = fields
        context['rels'] = []
        for rel in rels:
            for item in ['app', 'related_model']:
                rel[item] = rel[item].replace('_', '-')
            context['rels'].append(rel)

        context['ember_app'] = settings.FRONT_APPLICATION_NAME
        return context



from django.utils.module_loading import import_string
from rest_framework.serializers import PrimaryKeyRelatedField, ManyRelatedField, ModelSerializer

from export_app import settings


class ModelNotFoundException(Exception):

    def __init__(self, model, *args, **kwargs):
        self.model = model


class BaseSerializerExporter(object):
    """
    Returns the minimum info we need about a model
    """

    def __init__(self, *args, **kwargs):
        self.router = import_string(settings.ROUTER_PATH)
        super(BaseSerializerExporter, self).__init__(*args, **kwargs)

    def get_endpoint_for_basename(self, basename):
        if hasattr(self.router, '_endpoints') and basename in self.router._endpoints:
            return self.router._endpoints[basename]
        return None

    def get_viewset_for_basename(self, basename, with_endpoint=False):
        basename = basename.replace('_', '-')
        viewset = None
        endpoint = self.get_endpoint_for_basename(basename)
        if endpoint is not None:
            model_name = endpoint.singular_model_name
            application_name = endpoint.application_name
            viewset = endpoint.viewset
        else:
            for item in self.router.registry:
                if item[0] == basename:
                    viewset = item[1]
                    break
            if viewset is None:
                raise ModelNotFoundException(basename)

            model = viewset.serializer_class.Meta.model
            model_name = model._meta.model_name.lower()
            application_name = model._meta.app_label.lower()

        if with_endpoint:
            return viewset, model_name, application_name, endpoint
        return viewset, model_name, application_name

    def get_serializer_for_basename(self, basename):
        viewset, model_name, application_name, endpoint = \
            self.get_viewset_for_basename(basename, True)

        if endpoint is not None:
            model = endpoint.model
            serializer_instance = endpoint.serializer()
        else:
            model = viewset.serializer_class.Meta.model
            serializer_instance = viewset.serializer_class()

        return model, serializer_instance, model_name, application_name

    def get_fields_for_model(self, model, serializer_instance, adapter, target_app=None):
        return [], []


class SerializerExporterWithFields(BaseSerializerExporter):

    def get_fields_for_model(self, model, serializer_instance, adapter, target_app=None):

        fields = []
        relationships = []

        for field_name, field in serializer_instance.get_fields().items():

            if field_name == 'id':
                continue
            field_item = {
                'name': field_name,
                'type': adapter.field_type_mapping[field.__class__.__name__]
            }
            if isinstance(field, PrimaryKeyRelatedField) or isinstance(field, ManyRelatedField) \
                    or isinstance(field, ModelSerializer):
                if model is None:
                    field_item['related_model'] = field.queryset.model._meta.model_name.lower()
                    field_item['app'] = target_app if target_app is not None else \
                        field.queryset.model._meta.app_label.lower()
                    relationships.append(field_item)
                else:
                    model_field = model._meta.get_field(field_name)
                    field_item['related_model'] = model_field.related_model._meta.model_name.lower()
                    field_item['app'] = target_app if target_app is not None else \
                        model_field.related_model._meta.app_label.lower()
                    relationships.append(field_item)
                    if hasattr(model_field, 'field'):
                        field_item['inverse']=model_field.field.name
                    elif hasattr(model_field, 'remote_field') and \
                            getattr(model_field.remote_field, 'related_name', None) is not None:
                        field_item['inverse']=model_field.remote_field.related_name
                    if field_item.get('inverse', '-')[-1] == '+':
                        field_item.pop('inverse')
                if isinstance(field, ModelSerializer):
                    if field.many:
                        field_item[type] = adapter.field_type_mapping['ManyRelatedField']
                    else:
                        field_item[type] = adapter.field_type_mapping['PrimaryKeyRelatedField']
            else:
                fields.append(field_item)

        return fields, relationships
